#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import hashlib
import base64

from seguid._manip import reverse
from seguid._manip import rotate
from seguid._manip import rotate_to_min

from seguid._tables import tablefactory
from seguid._asserts import assert_in_alphabet
from seguid._asserts import assert_complementary

seguid_prefix: str = "seguid="
lsseguid_prefix: str = "lsseguid="
csseguid_prefix: str = "csseguid="
ldseguid_prefix: str = "ldseguid="
cdseguid_prefix: str = "cdseguid="
b64alphabet = set("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/_-")
short = 6


def _seguid(
    seq: str,
    alphabet: str = "{DNA}",
    encoding: callable = base64.standard_b64encode,
) -> str:
    assert callable(encoding)
    assert seq, "A sequence must not be empty"
    assert_in_alphabet(seq, alphabet=set(tablefactory(alphabet).keys()))
    m = hashlib.sha1()
    m.update(seq.encode("ASCII"))
    hs = encoding(m.digest())
    csum = f"{hs.decode('ASCII').rstrip('=')}"
    assert len(csum) == 27
    assert set(csum).issubset(b64alphabet)
    return csum


def _form(prefix, csum, form):
    longform = ""
    shortform = ""
    if form == "both":
        return csum[:short], prefix + csum
    elif form == "long":
        return prefix + csum
    if form == "short":
        return csum[:short]


# def _form(prefix, csum, form):
#     longform = ""
#     shortform = ""
#     if form != "short":
#         longform = prefix + csum
#     if form != "long":
#         shortform = csum[:short]
#     return " ".join((shortform, longform)).strip()


def seguid(seq: str, alphabet: str = "{DNA}", form: str = "long") -> str:
    """SEGUID v1 checksum for linear protein or single-stranded DNA.

    .. warning::
        ``seguid()`` (obsolete) is superseded by :func:`lsseguid()` (recommended).

    Given a nucleotide or amino-acid sequence ``seq`` in uppercase, the function returns
    a string containing the **SE**quence **G**lobally **U**nique **ID**entifier (**SEGUID**).
    The SEGUID is defined as the Base64 encoded SHA1 checksum calculated for the sequence
    in uppercase with the trailing padding symbol (``=``) removed.

    The original definition of the SEGUID v1 checksum algorithm (Babnigg & Giometti, 2006)
    included transformation to uppercase before calculating the checksum.
    Here, ``seguid()`` does *not* coerce the input sequence to upper case. If your input sequence
    has lower-case symbols, you can use :meth:`str.upper` to achieve what the original method does.
    ``seguid()`` only accepts symbols as specified by the `alphabet` argument.
    Thus, our implementation is more conservative, which has the benefit of
    lowering the risk of passing the incorrect sequence by mistake.

    The resulting checksum string may contain forward slash (``/``) and plus-sign (``+``) symbols.
    These characters cannot be a part of a Uniform Resource Locator (URL) or a filename on
    some operating systems. The SEGUID v2 checksum produced by :func:`lsseguid()` is similar to the
    SEGUID v1 checksum by ``seguid()``, but uses the Base64url encoding that do not produce
    these characters.

    The checksum is prefixed with ``seguid=``.

    Examples
    --------
    >>> seguid("AT")
    'seguid=Ax/RG6hzSrMEEWoCO1IWMGska+4'

    """
    return _form(
        seguid_prefix,
        _seguid(seq, alphabet=alphabet, encoding=base64.standard_b64encode),
        form,
    )


def lsseguid(seq: str, alphabet: str = "{DNA}", form: str = "long") -> str:
    """SEGUID checksum for linear single-stranded DNA.

    The same as the :func:`seguid()` function except that forward slashes (``/``) and plus signs (``+``)
    in the resulting checksum are replaced by underscores (``_``) and minus signs (``-``), respectively
    following the Base64url standard in RFC 4648.

    This checksum is applicable to linear single-stranded DNA
    sequences and protein sequences, among other sequences.  If
    protein sequences are analyzed, the alphabet argument should be
    ``"{protein}"`` or ``"{protein-extended}"``.

    The checksum is prefixed with ``lsseguid=``.

    Examples
    --------
    >>> lsseguid("AT")
    'lsseguid=Ax_RG6hzSrMEEWoCO1IWMGska-4'

    """
    return _form(
        lsseguid_prefix,
        _seguid(seq, alphabet=alphabet, encoding=base64.urlsafe_b64encode),
        form,
    )


def csseguid(seq: str, alphabet: str = "{DNA}", form: str = "long") -> str:
    r"""SEGUID checksum for circular single-stranded DNA.

    The ``csseguid()`` is the :func:`lsseguid()` checksum calculated
    for the lexicographically smallest string rotation of ``seq``.
    This checksum is Only defined for circular single-stranded
    sequences.

    The checksum is prefixed with ``csseguid=``.

    Examples
    --------
    >>> csseguid("ATTT")
    'csseguid=ot6JPLeAeMmfztW1736Kc6DAqlo'
    >>> lsseguid("ATTT")
    'lsseguid=ot6JPLeAeMmfztW1736Kc6DAqlo'
    >>> csseguid("TTTA")
    'csseguid=ot6JPLeAeMmfztW1736Kc6DAqlo'
    >>> lsseguid("TTTA")
    'lsseguid=8zCvKwyQAEsbPtC4yTV-pY0H93Q'

    """
    return _form(
        csseguid_prefix,
        _seguid(
            rotate_to_min(seq), alphabet=alphabet, encoding=base64.urlsafe_b64encode
        ),
        form,
    )


def ldseguid(
    watson: str, crick: str, alphabet: str = "{DNA}", form: str = "long"
) -> str:
    r"""SEGUID checksum for linear double-stranded DNA.

    This function calculates the SEGUID checksum for a double-stranded DNA (dsDNA) sequence defined by two
    strings representing the upper (Watson) and the complementary (Crick) DNA strands. Watson and Crick
    strands should be of equal length.
    Optional single-stranded DNA regions in the ends are indicated by a dash (``-``) in either strand.

    The algorithm first selects the lexicographically smallest of the
    Watson and Crick strands.  The two string are then joined 5'-3',
    separated by a semicolon (``;``), and the :func:`lsseguid()`
    function is used on the resulting string.

    For example, consider the linear dsDNA sequence defined by ``watson="-TATGCC"`` and ``crick="-GCATAC"`` as in:

    ::

        dsDNA    SEGUID checksum

        -TATGCC  ldseguid=rr65d6AYuP-CdMaVmdw3L9FPt6I
         |||||
        CATACG-

        -GCATAC  ldseguid=rr65d6AYuP-CdMaVmdw3L9FPt6I
         |||||
        CCGTAT-

    The SEGUID algorithm identifies the ``"-GCATAC"`` strand as the lexicographic smallest of the two. Then it concattenates the two as:
    ::

        "-GCATAC" + ";" + "-TATGCC"

    and calculates the final checksum based on that sequence.

    The checksum is prefixed with ``ldseguid=``.

    Examples
    --------
    >>> ldseguid("-TATGCC", "-GCATAC")
    'ldseguid=rr65d6AYuP-CdMaVmdw3L9FPt6I'
    >>> ldseguid("-GCATAC", "-TATGCC")
    'ldseguid=rr65d6AYuP-CdMaVmdw3L9FPt6I'

    """
    assert watson, "Watson sequence must not be empty"
    assert crick, "Crick sequence must not be empty"
    assert len(watson) == len(crick)
    assert_complementary(watson, crick, alphabet=alphabet)

    tb = tablefactory(alphabet)
    assert len(set(tb.values())) > 1, "Was a single-stranded alphabet used by mistake?"

    exalphabet = alphabet + ",--,;;"

    if watson < crick:
        spec = watson + ";" + crick
    else:
        spec = crick + ";" + watson

    return _form(
        ldseguid_prefix,
        _seguid(spec, alphabet=exalphabet, encoding=base64.urlsafe_b64encode),
        form,
    )


def cdseguid(
    watson: str, crick: str, alphabet: str = "{DNA}", form: str = "long"
) -> str:
    """SEGUID checksum for circular double-stranded DNA.

    The ``cdseguid()`` is the :func:`lsseguid()` checksum calculated for the lexicographically
    smallest string rotation of a double-stranded DNA sequence. Only defined for circular
    sequences.

    The checksum is prefixed with ``cdseguid=``.
    """
    from seguid._config import _min_rotation

    assert watson, "Watson sequence must not be empty"
    assert crick, "Crick sequence must not be empty"
    assert len(watson) == len(crick)
    assert_complementary(watson, crick, alphabet=alphabet)

    amount_watson = _min_rotation(watson)
    watson_min = rotate(watson, amount=amount_watson)
    amount_crick = _min_rotation(crick)
    crick_min = rotate(crick, amount=amount_crick)

    # Keep the "minimum" of the two variants
    if watson_min < crick_min:
        w = watson_min
        c = rotate(crick, amount=-amount_watson)
    else:
        w = crick_min
        c = rotate(watson, amount=-amount_crick)

    return _form(
        cdseguid_prefix,
        ldseguid(watson=w, crick=c, alphabet=alphabet, form="long")[
            len(ldseguid_prefix) :
        ],
        form,
    )



from seguid._manip import reverse
import re


def is_staggered(watson, crick):
    return "-" in watson or "-" in crick


def escape_sequence_spec(spec):
    spec = spec.replace("\n", "\\n")
    return spec


def parse_sequence_string(spec: str) -> str:
    """docstring."""
    assert isinstance(spec, str)

    ## Single- or double-stranded sequence?
    pattern = r"^([0-9A-Za-z-]+)$"
    if re.match(pattern, spec):
        return "ss", spec

    ## "<watson>;<crick>"?
    pattern = r"^([0-9A-Za-z-]+);([0-9A-Za-z-]+)$"
    match = re.match(pattern, spec)
    if match:
        watson = match.group(1)
        crick = match.group(2)
        rcrick = reverse(crick)
        if len(watson) != len(crick):
            raise ValueError(
                "Double-strand sequence string specifies two strands of different lengths (%d != %d): '%s'"
                % (len(watson), len(crick), spec)
            )

        if is_staggered(watson, crick):
            ## Watson and reverse Crick must not be staggered on the same side
            if watson.startswith("-") and rcrick.startswith("-"):
                raise ValueError(
                    "Please trim the staggering. Watson and Crick are both staggered at the beginning of the double-stranded sequence: '%s'"
                    % spec
                )
            if watson.endswith("-") and rcrick.endswith("-"):
                raise ValueError(
                    "Please trim the staggering. Watson and Crick are both staggered at the end of the double-stranded sequence: '%s'"
                    % spec
                )

        return "ds", spec, watson, crick

    ## "<watson>\n<rev crick>"?
    pattern = r"^([0-9A-Za-z-]+)\n([0-9A-Za-z-]+)$"
    match = re.match(pattern, spec)
    if match:
        watson = match.group(1)
        rcrick = match.group(2)
        crick = reverse(rcrick)
        if len(watson) != len(crick):
            raise ValueError(
                "Double-strand sequence string specifies two strands of different lengths (%d != %d): '%s'"
                % (len(watson), len(crick), escape_sequence_spec(spec))
            )

        if is_staggered(watson, crick):
            ## Watson and reverse Crick must not be staggered on the same side
            if watson.startswith("-") and rcrick.startswith("-"):
                raise ValueError(
                    "Please trim the staggering. Watson and Crick are both staggered at the beginning of the double-stranded sequence: '%s'"
                    % escape_sequence_spec(spec)
                )
            if watson.endswith("-") and rcrick.endswith("-"):
                raise ValueError(
                    "Please trim the staggering. Watson and Crick are both staggered at the end of the double-stranded sequence: '%s'"
                    % escape_sequence_spec(spec)
                )

        return "ds", spec, watson, crick

    raise ValueError(
        "Syntax error in sequence string: '%s'" % escape_sequence_spec(spec)
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from seguid._tables import tablefactory
import seguid._manip
from string import ascii_letters
from string import digits
import re


def assert_in_alphabet(seq: str, alphabet: set):

    assert isinstance(seq, str), "Argument 'seq' must be an string"
    assert isinstance(alphabet, set), "Argument 'alphabet' must be a set"
    assert len(alphabet) > 0, "Argument 'alphabet' must not be empty"
    assert not alphabet - (
        set(ascii_letters) | set(digits) | set("-\n;")
    ), "Only A-Z a-z allowed"
    first = list(alphabet)[0]
    assert isinstance(first, str), "Argument 'alphabet' should contain 'str' elements"

    # Nothing to do?
    if len(seq) == 0:
        return

    unknown = set(c for c in seq if c not in (k for k in alphabet))

    if unknown:
        missing = " ".join(unknown)
        missing = repr(missing)  ## escape '\n', '\t', ...
        raise ValueError("Detected symbols " f"{missing} not in the 'alphabet'")


def assert_alphabet(alphabet: dict):
    assert isinstance(alphabet, dict), "Argument 'alphabet' must be a dict"
    keys = alphabet.keys()
    values = alphabet.values()
    if isinstance(next(iter(values)), str):
        values2 = set()
        for value in values:
            for ch in value:
                values2.add(ch)
        values = values2

    # Assert that the set of values are also in the set of keys
    unknown = set(str(k) for k in keys if str(k) not in (str(v) for v in values))

    if unknown:
        missing = " ".join(unknown)
        missing = repr(missing)  ## escape '\n', '\t', ...
        raise ValueError(
            f"Detected keys ({missing}) in 'alphabet' that are not in the values"
        )


# def assert_checksum(checksum):
#     checksum = "cdseguid:AWD-dt5-TEua8RbOWfnctJIu9nA"
#     mobj = re.match("(?:|sl|sc|ds|dc)seguid:(.+)", checksum)
#     assert len(mobj.group(1)) == 27
#     b64 = set('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/_-')
#     assert set(mobj.group(1)).issubset(b64)


def reverse(seq) -> str:
    """Reverses a DNA sequence"""
    assert isinstance(seq, str), "Argument 'seq' must be an string"
    return seq[::-1]


def assert_complementary(
    watson: str, crick: str, alphabet: dict
):
    ## Validate 'alphabet':
    tb = tablefactory(alphabet)
    keys = tb.keys()
    values = tb.values()

    ## FIXME: Does this really work? See next line.
    assert len(tb.values()) > 1, "Was a single-stranded alphabet used by mistake?"
    assert (
        len(next(iter(values))) == 1
    ), "Was a single-stranded alphabet used by mistake?"
    assert_alphabet(tb)

    if not "-" in keys:
        tb["-"] = "-"
        keys = tb.keys()
        values = tb.values()

    ## Validate 'watson' and 'crick':
    assert len(watson) == len(crick)
    assert_in_alphabet(watson, alphabet=set(keys))
    assert_in_alphabet(crick, alphabet=set(keys))
    crick = reverse(crick)

    for kk in range(0, len(watson)):
        #        print(str(kk) + ": " + watson[kk] + "-" + crick[kk])
        if watson[kk] == "-" or crick[kk] == "-":
            continue
        set_kk = tb[watson[kk]]
        if not crick[kk] in set_kk:
            raise ValueError(
                "Non-complementary basepair (%s,%s) detected at position %d"
                % (watson[kk], crick[kk], kk + 1)
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import array
# The 'u' format code (wchar_t) has been deprecated
# https://docs.python.org/3/deprecations/index.html#pending-removal-in-python-3-16
typecode = "w" if "w" in array.typecodes else "u"

def rotate(seq: str, amount: int = 0) -> str:
    """Rotates a circular, DNA sequence a certain amount.

    Rotates sequence 'seq', 'amount' number of symbols to the right.
    A rotation 'amount' is the same as a rotation 'amount + n * len(seq)'
    for any integer 'n'.

    Returns the rotated sequence as a string of length 'len(seq)'.

    Examples
    --------
    >>> seq = "ABCDEFGH"
    >>> rotate(seq, 0)
    'ABCDEFGH'

    >>> rotate(seq, +1)
    'BCDEFGHA'

    >>> rotate(seq, +7)
    'HABCDEFG'

    >>> rotate(seq, -1)
    'HABCDEFG'

    >>> rotate(seq, +8)
    'ABCDEFGH'
    """
    assert isinstance(seq, str), "Argument 'seq' must be an string"
    assert isinstance(amount, int), "Argument 'amount' must be an integer"

    ## Nothing to rotate?
    if len(seq) == 0:
        return seq

    amount = amount % len(seq)

    ## Rotate?
    if amount > 0:
        seq = seq[amount:] + seq[:amount]

    return seq


def reverse(seq) -> str:
    """Reverses a DNA sequence"""
    assert isinstance(seq, str), "Argument 'seq' must be an string"
    return seq[::-1]


def min_rotation_py(s: str) -> int:
    """Start position for the smallest rotation of a string s (pure Python).

    Algorithm described in:

    Pierre Duval, Jean. 1983. Factorizing Words
    over an Ordered Alphabet. Journal of Algorithms & Computational Technology
    4 (4) (December 1): 363–381. and Algorithms on strings and sequences based
    on Lyndon words, David Eppstein 2011.
    https://gist.github.com/dvberkel/1950267

    This is a pure python implementation, considerably slower than the
    min_rotation function from pydivsufsort that is imported at the top of
    this script.

    Should only be used if pydivsufsort.min_rotation can not be used.

    Note that both functions are case-sensitive and sorts by ASCII-code order
    or "ASCIIbetical" order so:

    - Uppercase come before lowercase letters; for example, "Z" precedes "a"
    - Digits and several punctuation marks come before letters.

    See the last two examples below for an example of the consequences of this.

    Examples
    --------
    >>> min_rotation_py("TAAA")
    1
    >>> "TAAA"[1:] + "TAAA"[:1]
    'AAAT'
    >>> s = "ACAACAAACAACACAAACAAACACAAC"
    >>> min_rotation_py(s)
    14
    >>> s[14:] + s[:14]
    'AAACAAACACAACACAACAAACAACAC'
    """

    prev, rep = None, 0
    ds = array.array(typecode, 2 * s)
    lens = len(s)
    lends = lens * 2
    old = 0
    k = 0
    w = ""
    while k < lends:
        i, j = k, k + 1
        while j < lends and ds[i] <= ds[j]:
            i = (ds[i] == ds[j]) and i + 1 or k
            j += 1
        while k < i + 1:
            k += j - i
            prev = w
            w = ds[old:k]
            old = k
            if w == prev:
                rep += 1
            else:
                prev, rep = w, 1
            if len(w) * rep == lens:
                return old - i
    return 0


def rotate_to_min(s: str) -> int:
    from seguid._config import _min_rotation

    ## Assert upper-case letters are ordered before lower-case letters
    assert _min_rotation("Aa") == 0

    amount = _min_rotation(s)
    s = rotate(s, amount=amount)
    return s


# def linearize_circular_dsDNA(watson, crick, position):
#     cposition = len(watson) - position
#     swatson = watson[position:] + watson[:position]
#     scrick = watson[cposition:] + watson[:cposition]
#     return (swatson, scrick, 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

alphabets = {
               "{DNA}": "GC,AT",
               "{RNA}": "GC,AU",
      "{DNA-extended}": "GC,AT,BV,DH,KM,SS,RY,WW,NN",
      "{RNA-extended}": "GC,AU,BV,DH,KM,SS,RY,WW,NN",
           "{protein}": "A,C,D,E,F,G,H,I,K,L,M,N,P,Q,R,S,T,V,W,Y,O,U",
  "{protein-extended}": "A,C,D,E,F,G,H,I,K,L,M,N,P,Q,R,S,T,V,W,Y,O,U,B,J,X,Z"
}

def tablefactory(argument: str):
    # argument = "{protein},X,Z"
    # argument = "{DNA},BV,DH,KM,NN,SS,WW"
    # argument = "AT,CG"
    # argument = 'A,C,D,E,F,G,H,I,K,L,M,N,P,Q,R,S,T,V,W,Y'
    for name in alphabets.keys():
        argument = argument.replace(name, alphabets[name])

    n_expected=-1
    alphabet = {}
    for spec in argument.split(","):
        n=len(spec)
        if n_expected < 0:
            n_expected = n
        else:
            assert n == n_expected
        if n == 1:
            alphabet[spec] = ""
        elif n == 2:
            if spec[0] in alphabet.keys():
                alphabet[spec[0]] = alphabet[spec[0]] + spec[1]
            else:
                alphabet[spec[0]] = spec[1]
            if spec[1] in alphabet.keys():
                alphabet[spec[1]] = alphabet[spec[1]] + spec[0]
            else:
                alphabet[spec[1]] = spec[0]
        else:
            raise ValueError("Unknown alphabet specification: %s" % spec)
        
    return alphabet



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import warnings
from seguid._manip import min_rotation_py

## The min_rotation() function
_min_rotation = min_rotation_py


def set_min_rotation(which):
    assert isinstance(which, str), "Argument 'which' must be an string"
    if which == "built-in":
        _min_rotation = min_rotation_py
    elif which == "pydivsufsort":
        from pydivsufsort import min_rotation as mr

        def min_rotation_pydivsufsort(s):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
            result = mr(s)
            return result

        _min_rotation = min_rotation_pydivsufsort
    else:
        raise ValueError("Argument 'which' should be either 'built-in' or '': " + which)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SEGUID checksums for linear, circular, single- and double-stranded sequences
----------------------------------------------------------------------------

This package provides four functions, ``lsseguid()``, ``csseguid()``, ``ldseguid()``, and ``cdseguid()`` (Table 1), for calculating SEGUID v2 checksums, and one function, ``seguid()``, for calculating SEGUID v1 checksums.  SEGUID v2 is described in Pereira et al. (2024), and SEGUID v1 in Babnigg & Giometti (2006).

================ ============== ====================
 Topology         Strandedness   Function
================ ============== ====================
 **l**inear     **s**ingle     ``lsseguid()``
 **c**ircular   **s**ingle     ``csseguid()``
 **l**inear     **d**ouble     ``ldseguid()``
 **c**ircular   **d**ouble     ``cdseguid()``
================ ============== ====================

*Table 1: The four functions that calculate SEGUID v2 checksums for biological sequences with linear or circular topology and a single or two complementary strands.*

This package works without external dependencies, but ``csseguid()`` and ``cdseguid()``
can be made faster by installing `pydivsufsort <https://pypi.org/project/pydivsufsort/>`_, provides a faster implementation of the algorithm for finding the
`smallest string rotation <https://en.wikipedia.org/wiki/Lexicographically_minimal_string_rotation>`_.


Usage
-----

.. code-block:: python

    lsseguid(seq, alphabet="{DNA}", form="long")

    csseguid(seq, alphabet="{DNA}", form="long")

    ldseguid(watson, crick, alphabet="{DNA}", form="long")

    cdseguid(watson, crick, alphabet="{DNA}", form="long")

    seguid(seq, alphabet="{DNA}", form="long")


Function arguments
------------------

**seq** (string)
The sequence for which the checksum should be calculated. The sequence may only comprise of symbols in the
alphabet specified by the alphabet argument.

**watson, crick** (strings)
Two reverse-complementary DNA sequences. Both sequences should be specified in the 5'-to-3' direction.

**alphabet** (string)
The type of sequence used. If ``"{DNA}"`` (default), then the input is a DNA sequence. If ``"{RNA}"``, then the
input is an RNA sequence. If ``"{protein}"``, then the input is an amino-acid sequence. If ``"{DNA-extended}"``
or ``"{RNA-extended}"``, then the input is a DNA or RNA sequence specified an extended set of symbols, including
IUPAC symbols (4). If ``"{protein-extended}"``, then the input is an amino-acid sequence with an extended set
of symbols, including IUPAC symbols (5). A custom alphabet may also be used. A non-complementary alphabet
is specified as a comma-separated set of single symbols, e.g. ``"X,Y,Z"``. A complementary alphabet is specified
as a comma-separated set of paired symbols, e.g. ``"AT,CG"``. It is also possible to extend a pre-defined
alphabet, e.g. ``"{DNA},XY"``.

**form** (string)
How the checksum is presented. If ``"long"`` (default), the full-length checksum is returned. If ``"short"``,
the short, six-digit checksum is returned. If ``"both"``, both the short and the long checksums are returned.


Value
-----

The SEGUID functions return a single string, if form is either ``"long"`` or ``"short"``. If form
is ``"both"``, then a tuple of two strings is returned, where the first component holds the ``"short"``
checksum and the second the ``"long"`` checksum. The long checksum, without the prefix, is a string with 27
characters. The short checksum, without the prefix, is the first six characters of the long checksum.
All long checksums are prefixed with a label indicating which SEGUID method was used.

All functions produce checksums using the Base64url encoding ("Base 64 Encoding with URL and Filename Safe Alphabet"), with the exception for ``seguid()``,
which uses Base64 encoding. The "long" checksums returned are always 27-character long. This is because the SHA-1 hash (6) is 160-bit long
(20 bytes), which result in the encoded representation always end with a padding character (``=``) so that the
length is a multiple of four character. We relax this requirement, by dropping the padding character.


Base64 and Base64url encodings
------------------------------

The Base64url encoding is the Base64 encoding with non-URL-safe characters substituted with URL-safe ones (Josefsson, 2006).
Specifically, the plus symbol (``+``) is replaced by the minus symbol (``-``), and the forward slash (``/``) is replaced
by the underscore symbol (``_``).

The Base64 checksum used for the original SEGUID checksum is not guaranteed to contain symbols
that can safely be used as-is in a Uniform Resource Locator (URL). Specifically, it may consist of forward
slashes (``/``) and plus symbols (``+``), which are characters that carry special meaning in a URL. For the same
reason, a Base64 checksum cannot safely be used as a file or directory name, because it may have a forward slash.





References
----------

1. G Babnigg & CS Giometti, A database of unique protein sequence identifiers for proteome studies. Proteomics. 2006 Aug;6(16):4514-22, `doi:10.1002/pmic.200600032 <https://doi.org/10.1002/pmic.200600032>`_.

2. H Pereira, PC Silva, WM Davis, L Abraham, G Babnigg, H Bengtsson & B Johansson, SEGUID v2: Extending SEGUID Checksums for Circular, Linear, Single- and Double-Stranded Biological Sequences, bioRxiv, `doi:10.1101/2024.02.28.582384 <https://doi.org/10.1101/2024.02.28.582384>`_.

3. S Josefsson, The Base16, Base32, and Base64 Data Encodings, RFC 4648, October 2006, `doi:10.17487/RFC4648 <https://doi.org/10.17487/RFC4648>`_.

4. Wikipedia article 'Nucleic acid notation', February 2024, https://en.wikipedia.org/wiki/Nucleic_acid_notation.

5. Wikipedia article 'Amino acids', February 2024, https://en.wikipedia.org/wiki/Amino_acid.

6. Wikipedia article 'SHA-1' (Secure Hash Algorithm 1), December 2023, https://en.wikipedia.org/wiki/SHA-1.

"""

try:
    # importlib only avaliable in Python 3.8 and up
    from importlib.metadata import version as _version
except ModuleNotFoundError:
    try:
        # pkg_resources avaliable in Python 3.6-3.7 from setuptools =< 59.6.0
        import pkg_resources
    except ModuleNotFoundError:
        __version__ = "0.0.0"
    else:
        __version__ = pkg_resources.get_distribution(__package__).version
else:
    __version__ = _version(__package__)

from seguid._chksum import seguid
from seguid._chksum import lsseguid
from seguid._chksum import csseguid
from seguid._chksum import ldseguid
from seguid._chksum import cdseguid

__all__ = ["seguid", "lsseguid", "csseguid", "ldseguid", "cdseguid"]



from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from seguid import __version__

from seguid._chksum import seguid
from seguid._chksum import lsseguid
from seguid._chksum import csseguid
from seguid._chksum import ldseguid
from seguid._chksum import cdseguid
from seguid._manip import reverse
from seguid._reprutils import parse_sequence_string

parser = ArgumentParser(
    prog="python -m seguid",
    description="seguid: Sequence Globally Unique Identifier Checksums for Linear, Circular, Single- and Double-Stranded Biological Sequences",
    epilog=
"""
Predefined alphabets:
 '{DNA}'              Complementary DNA symbols (= 'AT,CG')
 '{DNA-extended}'     Extended DNA (= '{DNA},BV,DH,KM,SS,RY,WW,NN')
 '{RNA}'              Complementary RNA symbols (= 'AU,CG')
 '{RNA-extended}'     Extended DNA (= '{RNA},BV,DH,KM,SS,RY,WW,NN')
 '{protein}'          Amino-acid symbols (= 'A,C,D,E,F,G,H,I,K,L,M,N,P,Q,R,S,T,V,W,Y')
 '{protein-extended}' Amino-acid symbols (= '{protein},O,U,B,J,Z,X')

Examples:
python -m seguid --version
python -m seguid --help

echo 'ACGT' | python -m seguid --type=lsseguid
python -m seguid --type=lsseguid <<< 'ACGT'
python -m seguid --type=cdseguid <<< $'ACGT\\nTGCA' # two-line syntax
python -m seguid --type=ldseguid <<< 'ACGT;ACGT'   # watson-crick syntax
python -m seguid --type=ldseguid <<< $'-CGT\\nTGCA' # two-line syntax
python -m seguid --type=ldseguid <<< '-CGT;ACGT'   # watson-crick syntax
python -m seguid --type=lsseguid --alphabet='{RNA}' <<< 'ACGU'

Version: """ + __version__ + """
Copyright: Bjorn Johansson, Henrik Bengtsson (2023-2024)
License: MIT
""",
    formatter_class=RawDescriptionHelpFormatter  # Keeps the formatting of the epilog
)

parser.add_argument("--version",  action="store_true", help="Display version")
parser.add_argument("--type",     type=str, nargs="?", help="Type of checksum to calculate (lsseguid, csseguid, ldseguid, cdseguid, or seguid [default])")
parser.add_argument("--alphabet", type=str, nargs="?", help="Set of symbols for the input sequence, e.g. '{DNA}', 'AC,GT', and '{DNA},AU'")
parser.add_argument("--form",     type=str, nargs="?", help="Form of checksum to display ('long' [default], 'short', or 'both')")


args = vars(parser.parse_args())

if args.pop("version"):
    print(__version__)
else:
    form = args.pop("form")
    if form == None:
        form = "long"

    type = args.pop("type")
    if type == None:
        type = "seguid"

    alphabet = args.pop("alphabet")
    if alphabet == None:
        alphabet = "{DNA}"

    ## Read sequence data from the standard input
    lines = []
    try:
        while True:
            line = input()
            if not line:  # Optionally, break if an empty line is encountered
                break
            lines.append(line)
    except EOFError:
        pass
    seq = "\n".join(lines)

    if type == "seguid":
        res = seguid(seq, alphabet=alphabet, form=form)
    elif type == "lsseguid":
        res = lsseguid(seq, alphabet=alphabet, form=form)
    elif type == "csseguid":
        res = csseguid(seq, alphabet=alphabet, form=form)
    elif type == "ldseguid":
        void, void, watson, crick = parse_sequence_string(seq)
        res = ldseguid(watson=watson, crick=crick, alphabet=alphabet, form=form)
    elif type == "cdseguid":
        void, void, watson, crick = parse_sequence_string(seq)
        res = cdseguid(watson=watson, crick=crick, alphabet=alphabet, form=form)
    else:
        raise ValueError("Unknown --type='" + type + "'")

    if isinstance(res, tuple):
        res = " ".join(res)
        
    print(res)

